"""add per channel TTL

Revision ID: cd404ed93cc0
Revises: 303ff70c27fc
Create Date: 2021-01-07 20:12:23.527311

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = 'cd404ed93cc0'
down_revision = '303ff70c27fc'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        'channels',
        sa.Column('ttl', sa.Integer(), server_default='36000', nullable=False),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('channels', 'ttl')
    # ### end Alembic commands ###
