"""add channel metadata

Revision ID: 3c3288034362
Revises: ea6eba9a9ffc
Create Date: 2021-02-09 18:45:13.198051

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '3c3288034362'
down_revision = 'ea6eba9a9ffc'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        'channels',
        sa.Column('channel_metadata', sa.String(), server_default='{}', nullable=False),
    )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    op.drop_column('channels', 'channel_metadata')

    # ### end Alembic commands ###
