"""scheduling spec for jobs

Revision ID: cddba8e6e639
Revises: 0a0ab48887ab
Create Date: 2021-03-02 11:26:04.330580

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = 'cddba8e6e639'
down_revision = '0a0ab48887ab'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('jobs', schema=None) as batch_op:
        batch_op.add_column(
            sa.Column('repeat_every_seconds', sa.Integer(), nullable=True)
        )
        batch_op.add_column(sa.Column('start_at', sa.DateTime(), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('jobs', schema=None) as batch_op:
        batch_op.drop_column('start_at')
        batch_op.drop_column('repeat_every_seconds')

    # ### end Alembic commands ###
